import resonances
import pandas as pd
from pathlib import Path
import os


class Matrix:
    catalog_file = ''
    matrix = None
    planets = None

    @classmethod
    def dump(cls):
        if cls.matrix is None:
            cls.build()

        cls.matrix.to_csv(cls.catalog_full_filename())

    @classmethod
    def catalog_full_filename(cls) -> str:
        catalog_file = f"{os.getcwd()}/{resonances.config.get(cls.catalog_file)}"
        return catalog_file

    @classmethod
    def load(cls, reload=False):
        catalog_file = Path(cls.catalog_full_filename())
        if (not catalog_file.exists()) or (reload):
            cls.dump()
        else:
            catalog = pd.read_csv(catalog_file)
            cls.matrix = catalog



import rebound

# PLANETS_LONGITUDE = {
#     "Mercury": 0.0,
#     "Venus": 7.452,
#     "Earth": 17.368,
#     "Mars": 17.916,
#     "Jupiter": 4.257,
#     "Saturn": 28.243,
#     "Uranus": 0.0,
#     "Neptune": 0.0,
# }

PLANETS_AXIS = {
    "Mercury": 0.38709843,
    "Venus": 0.72332102,
    "Earth": 1.00000018,
    "Mars": 1.52371243,
    "Jupiter": 5.20248019,
    "Saturn": 9.54149883,
    "Uranus": 19.18797948,
    "Neptune": 30.06952752,
    "Pluto": 39.48686035,
}

masses = rebound.units.masses_SI

PLANETS_MASS = {
    "Mercury": masses['mmercury'] / masses['msun'],
    "Venus": masses['mvenus'] / masses['msun'],
    "Earth": masses['mearth'] / masses['msun'],
    "Mars": masses['mmars'] / masses['msun'],
    "Jupiter": masses['mjupiter'] / masses['msun'],
    "Saturn": masses['msaturn'] / masses['msun'],
    "Uranus": masses['muranus'] / masses['msun'],
    "Neptune": masses['mneptune'] / masses['msun'],
    "Pluto": masses['mpluto'] / masses['msun'],
}

K = 0.0172020989484
DAYS_IN_YEAR = 365.2422

SOLAR_SYSTEM = ["Mercury", "Venus", "Earth", "Mars", "Jupiter", "Saturn", "Uranus", "Neptune"]



import numpy as np
import pandas as pd
import rebound
from pathlib import Path
import os
from typing import List, Union

import resonances
import astdys
import datetime


class Simulation:
    def __init__(self, name=None):
        self.name = name
        if self.name is None:
            self.name = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        self.planets = self.list_of_planets()

        self.times = []
        self.bodies: List[resonances.Body] = []
        self.particles = []

        self.bodies_date = resonances.config.get('catalog.date')

        # Libration and filtering settings
        self.oscillations_cutoff = resonances.config.get('libration.oscillation.filter.cutoff')
        self.oscillations_filter_order = resonances.config.get('libration.oscillation.filter.order')

        self.periodogram_frequency_min = resonances.config.get('libration.periodogram.frequency.min')
        self.periodogram_frequency_max = resonances.config.get('libration.periodogram.frequency.max')
        self.periodogram_critical = resonances.config.get('libration.periodogram.critical')
        self.periodogram_soft = resonances.config.get('libration.periodogram.soft')

        self.libration_period_critical = resonances.config.get('libration.period.critical')
        self.libration_monotony_critical = resonances.config.get('libration.monotony.critical')
        self.libration_period_min = resonances.config.get('libration.period.min')

        self.sim = None

        self.Nout = None
        self.tmax = resonances.config.get('integration.tmax')
        self.integrator = resonances.config.get('integration.integrator')
        self.dt = resonances.config.get('integration.dt')
        if resonances.config.has('integration.integrator.safe_mode'):
            self.integrator_safe_mode = resonances.config.get('integration.integrator.safe_mode')
        else:  # pragma: no cover
            self.integrator_safe_mode = 1

        if resonances.config.has('integration.integrator.corrector'):
            self.integrator_corrector = resonances.config.get('integration.integrator.corrector')
        else:  # pragma: no cover
            self.integrator_corrector = None

        self.save = resonances.config.get('save')
        self.save_path = f"{resonances.config.get('save.path')}/{self.name}"
        self.save_summary = resonances.config.get('save.summary')

        self.plot = resonances.config.get('plot')
        self.plot_type = resonances.config.get('plot.type', 'both')
        self.plot_path = f"{resonances.config.get('plot.path')}/{self.name}/images"

        self.image_type = resonances.config.get('plot.image_type', 'png')
        self.data_source = resonances.config.get('data.source', 'astdys')

    def solar_system_full_filename(self) -> str:
        catalog_file = f"{os.getcwd()}/{resonances.config.get('solar_system_file')}"
        return catalog_file

    def create_solar_system(self, date: str = ''):
        solar_file = Path(self.solar_system_full_filename())
        if solar_file.exists():
            self.sim = rebound.Simulation(self.solar_system_full_filename())
        else:  # pragma: no cover
            self.sim = rebound.Simulation()
            if date != '':
                self.sim.add(self.list_of_planets(), date=date)
            elif self.data_source == 'astdys':
                self.sim.add(self.list_of_planets(), date=f"{astdys.catalog_time()} 00:00")  # date of AstDyS current catalogue
            else:
                self.sim.add(self.list_of_planets())
            self.sim.save(self.solar_system_full_filename())

    def add_body(self, elem_or_num, mmr: Union[str, resonances.MMR, List[resonances.MMR]], name='asteroid'):
        body = resonances.Body()

        if isinstance(mmr, str):
            mmr = [resonances.create_mmr(mmr)]

        if isinstance(mmr, resonances.MMR):
            mmr = [mmr]

        if len(mmr) == 0:
            raise Exception('You have to provide at least one resonance')

        if isinstance(elem_or_num, int) or (isinstance(elem_or_num, str)):
            elem = astdys.search(elem_or_num)
        elif isinstance(elem_or_num, dict):
            elem = elem_or_num
            if 'mass' in elem:
                body.mass = elem['mass']
        else:
            raise Exception('You can add body only by its number or all orbital elements')

        body.initial_data = elem
        body.name = name
        body.mmrs = mmr

        for elem in body.mmrs:
            elem.index_of_planets = self.get_index_of_planets(elem.planets_names)
        self.bodies.append(body)

    def add_bodies_to_simulation(self):
        for body in self.bodies:
            self.add_body_to_simulation(body)

    def add_body_to_simulation(self, body: resonances.Body):
        body.index_in_simulation = len(self.sim.particles)
        self.sim.add(
            m=body.mass,
            a=body.initial_data['a'],
            e=body.initial_data['e'],
            inc=body.initial_data['inc'],
            Omega=body.initial_data['Omega'],
            omega=body.initial_data['omega'],
            M=body.initial_data['M'],
            date=self.bodies_date,
            primary=self.sim.particles[0],
        )

    def setup_integrator(self, N_active=10):  # pragma: no cover
        self.sim.integrator = self.integrator
        self.sim.dt = self.dt
        self.sim.N_active = N_active

        if 'whfast' == self.integrator:
            self.sim.ri_whfast.safe_mode = 0
            if self.integrator_corrector is not None:
                self.sim.ri_whfast.corrector = self.integrator_corrector
        elif 'SABA' in self.integrator:
            self.sim.ri_saba.safe_mode = self.integrator_safe_mode

        self.sim.move_to_com()

    def run(self):
        self.add_bodies_to_simulation()
        for body in self.bodies:
            body.setup_vars_for_simulation(self.Nout)

        self.times = np.linspace(0.0, self.tmax, self.Nout)
        self.setup_integrator()

        ps = self.sim.particles

        for i, time in enumerate(self.times):
            self.sim.integrate(time)
            os = self.sim.calculate_orbits(primary=ps[0])

            for body in self.bodies:
                tmp = os[body.index_in_simulation - 1]  # ? -1 because Sun is not in os
                body.axis[i], body.ecc[i], body.longitude[i], body.varpi[i] = (
                    tmp.a,
                    tmp.e,
                    tmp.l,
                    tmp.Omega + tmp.omega,
                )
                for mmr in body.mmrs:
                    planets = []
                    for index in mmr.index_of_planets:
                        planets.append(os[index - 1])

                    body.angle(mmr)[i] = mmr.calc_angle(os[body.index_in_simulation - 1], planets)

        self.identify_librations()
        self.save_data()

    def save_data(self):
        if self.save_summary:
            self.save_simulation_summary()

        for body in self.bodies:
            for mmr in body.mmrs:
                if self.shall_save_body_in_mmr(body, mmr):
                    self.save_body_in_mmr(body, mmr)
                if self.shall_plot_body_in_mmr(body, mmr):
                    self.plot_body_in_mmr(body, mmr)

    def identify_librations(self):  # pragma: no cover
        for body in self.bodies:
            try:
                resonances.libration.body(self, body)
            except Exception as e:
                resonances.logger.error(f"Error identifying librations for {body.name}: {str(e)}")
                raise e

    def shall_save_body_in_mmr(self, body: resonances.Body, mmr: resonances.MMR):
        return self.process_status(body, mmr, self.save)

    def shall_plot_body_in_mmr(self, body: resonances.Body, mmr: resonances.MMR):
        return self.process_status(body, mmr, self.plot)

    def process_status(self, body: resonances.Body, mmr: resonances.MMR, variable) -> bool:
        if variable is None:
            return False

        if variable == 'all':
            return True

        if (variable == 'resonant') and (body.statuses[mmr.to_s()] > 0):
            return True

        if (variable == 'nonzero') and (body.statuses[mmr.to_s()] != 0):
            return True

        if (variable == 'candidates') and (body.statuses[mmr.to_s()] < 0):
            return True

        return False

    def plot_body_in_mmr(self, body: resonances.Body, mmr: resonances.MMR):
        self.check_or_create_save_path()
        resonances.resonance.plot.body(self, body, mmr, image_type=self.image_type)

    def save_body_in_mmr(self, body: resonances.Body, mmr: resonances.MMR):
        self.check_or_create_save_path()
        df_data = body.mmr_to_dict(mmr, self.times)
        if df_data is not None:
            df = pd.DataFrame(data=df_data)
            df.to_csv('{}/data-{}-{}.csv'.format(self.save_path, body.name, mmr.to_s()))

    def get_simulation_summary(self) -> pd.DataFrame:
        data = []
        for body in self.bodies:
            for mmr in body.mmrs:
                try:
                    s = ', '.join('({:.0f}, {:.0f})'.format(left, right) for left, right in body.periodogram_peaks_overlapping[mmr.to_s()])
                    data.append(
                        [
                            body.name,
                            mmr.to_s(),
                            body.statuses[mmr.to_s()],
                            body.libration_pure[mmr.to_s()],
                            body.libration_metrics[mmr.to_s()]['num_libration_periods'],
                            body.libration_metrics[mmr.to_s()]['max_libration_length'],
                            body.monotony[mmr.to_s()],
                            s,
                            body.initial_data['a'],
                            body.initial_data['e'],
                            body.initial_data['inc'],
                            body.initial_data['Omega'],
                            body.initial_data['omega'],
                            body.initial_data['M'],
                        ]
                    )
                except Exception as e:  # pragma: no cover
                    resonances.logger.error(f"Error getting summary for {body.name} and {mmr.to_s()}: {str(e)}.\n\n{str(body)}")

        df = pd.DataFrame(
            data,
            columns=[
                'name',
                'mmr',
                'status',
                'pure',
                'num_libration_periods',
                'max_libration_length',
                'monotony',
                'overlapping',
                'a',
                'e',
                'inc',
                'Omega',
                'omega',
                'M',
            ],
        )
        return df

    def save_simulation_summary(self) -> pd.DataFrame:
        self.check_or_create_save_path()
        df = self.get_simulation_summary()
        summary_filename = '{}/summary.csv'.format(self.save_path)
        summary_file = Path(summary_filename)
        if summary_file.exists():
            df.to_csv(summary_filename, mode='a', header=False, index=False)
        else:
            df.to_csv(summary_filename, mode='a', header=True, index=False)

        return df

    def check_or_create_save_path(self):
        Path(self.save_path).mkdir(parents=True, exist_ok=True)
        Path(self.plot_path).mkdir(parents=True, exist_ok=True)

    def list_of_planets(self):
        planets = [
            'Sun',
            'Mercury',
            'Venus',
            'Earth',
            'Mars',
            'Jupiter',
            'Saturn',
            'Uranus',
            'Neptune',
            'Pluto',
        ]

        return planets

    def get_index_of_planets(self, planets_names):
        arr = []
        for planet in planets_names:
            arr.append(self.planets.index(planet))
        return arr

    @property
    def tmax(self):
        return self.__tmax

    @tmax.setter
    def tmax(self, value):
        self.__tmax = value
        self.tmax_yrs = self.__tmax / (2 * np.pi)
        if self.Nout is None:
            self.Nout = int(self.__tmax / 100)

    @tmax.deleter
    def tmax(self):  # pragma: no cover
        del self.__tmax



import json
from pathlib import Path


def static_init(cls):
    if getattr(cls, "static_init", None):
        cls.static_init()
    return cls


@static_init
class config:
    config = None

    @classmethod
    def get(cls, key, default=None):
        try:
            value = cls.config[key]
        except Exception:
            if default is not None:
                return default
            raise Exception('There is no config with key = {}. The full config: {}'.format(key, json.dumps(cls.config)))
        return value

    @classmethod
    def has(cls, key):
        if key in cls.config:
            return True
        return False

    @classmethod
    def set(cls, key, value):
        if not cls.has(key):
            raise Exception('There is no config with key = {}. The full config: {}'.format(key, json.dumps(cls.config)))

        cls.config[key] = value

    @classmethod
    def static_init(cls):
        config_file_dir = Path(__file__).parent.resolve()
        config_file_path = '{}/config.json'.format(str(config_file_dir))
        config_file = Path(config_file_path)

        if not config_file.exists():  # pragma: no cover
            raise Exception('No config.json presented. Looking at {} Cannot continue working.'.format(config_file_path))

        with open(config_file_path, "r") as read_file:
            cls.config = json.load(read_file)



import math
import astdys
import resonances


def run(mmr: resonances.MMR, dump=100, max_iterations=1000, plot=False):
    df = astdys.search_by_axis(mmr.resonant_axis)
    asteroids = df['num'].tolist()

    num_particles = len(asteroids)
    num_iterations = int(math.ceil(num_particles / dump))
    for j in range(num_iterations):
        if j > (max_iterations - 1):
            resonances.logger.info(
                'Terminating because the app has reached the limit specified in max_iterations parameter ({}).'.format(max_iterations)
            )
            break

        sim = resonances.Simulation()
        sim.save, sim.save_path, sim.plot, sim.save_only_undetermined = False, 'cache/finder', plot, True
        sim.create_solar_system()

        # Find the number of calculations for a given step. It varies for the last one.
        if j < num_iterations - 1:
            num = dump
        else:
            num = num_particles % dump
            if num == 0:
                num = dump

        for i in range(num):
            key = j * dump + i
            sim.add_body(asteroids[key], mmr, '{}-{}'.format(asteroids[key], key))

        sim.run()
        resonances.logger.info('Iteration {} has finished. Processed from {}. Starting the new one.'.format(j, j * dump))



import re
import resonances


def create_mmr(coeff, planets_names=None):
    if isinstance(coeff, str):
        tmp = re.split('-|\\+', coeff)
        size = len(tmp)
        if 3 == size:
            return resonances.ThreeBody(coeff)
        elif 2 == size:
            return resonances.TwoBody(coeff)
        else:
            raise Exception(
                """Cannot create a resonance because the notation is wrong.
                 It should have either two or three bodies, i.e. 2J-1 or 4M-2J-1. Given {}.""".format(
                    coeff
                )
            )
    elif isinstance(coeff, list):
        size = len(coeff)
        if 6 == size:
            return resonances.ThreeBody(coeff, planets_names)
        elif 4 == size:
            return resonances.TwoBody(coeff, planets_names)
        else:
            raise Exception(
                'Cannot create a resonance because the number of coefficients is wrong. It should be equal to 2 or 3. Given {}.'.format(
                    len(coeff)
                )
            )

    raise Exception(
        'The argument should be either a string if you use the short notation (i.e. 2J-1) or a list containing 2 or 3 elements.'
    )



import numpy as np
from scipy import signal
from astropy.timeseries import LombScargle

import resonances.config


class libration:
    @classmethod
    def shift(cls, angle, distance=0):
        tmp = np.array(angle, copy=True)
        for i, elem in enumerate(tmp):
            if elem > np.pi:
                tmp[i] = tmp[i] - 2 * np.pi
        return tmp

    @classmethod
    def is_pure(cls, y):
        prev = y[0]
        num = 0
        for elem in y:
            num += 1
            if abs(elem - prev) > np.pi:
                return False
            prev = elem
        return True

    @classmethod
    def pure(cls, y):  # not working for librations that are not around 0 or +/- np.pi
        flag1 = cls.is_pure(y)
        if flag1:
            return True
        flag2 = cls.is_pure(cls.shift(y))
        if flag2:
            return True
        return False

    @classmethod
    def monotony_estimation(cls, data, crit=np.pi):
        num = 0
        prev = data[0]
        for elem in data:
            if prev - elem > crit:
                prev = elem
                continue
            if elem - prev > crit:
                num += 1
                prev = elem
                continue
            if elem < prev:
                num += 1  # num of "decreasing" points
            prev = elem
        return num / (len(data) - 1)

    @classmethod
    def find_breaks(cls, x, y, break_value=np.pi):
        """
        This function finds continuos breaks in a given dataset and return these breaks among with additional data.

        Parameters
        ----------

        x : list
            one-dimensional array of time
        y : list
            one-dimensional array of data where to find breaks
        break_value : float
            Used to determine a break. If the absolute diff between the previous one and
            the current is more than break_value, then there is a break. (Default: np.pi)

        Returns
        -------
        list
            Returns a multidimensional array:
            res[0] - the time when the break happened,
            res[1] - the direction of the break (1 if it intersects the top line and thus elem>prev, -1 otherwise),
            res[2] - values of the elements prior to the breaks,
            res[3] - the values in the break points.
        """
        prev = y[0]
        res = [[], [], [], []]  # break point, direction (1 or -1), prev, curr
        for i, elem in enumerate(y):
            if abs(elem - prev) > break_value:
                res[0].append(x[i])
                direction = 1 if elem > prev else -1
                res[1].append(direction)
                res[2].append(prev)
                res[3].append(elem)
            prev = elem
        return res

    @classmethod
    def circulation(cls, x, y):
        """
        Find and returns libration periods with some additional data.

        Parameters
        ----------

        x : list
            one-dimensional array of time
        y : list
            Input data to find breaks.

        Returns
        -------

        list
            Returns a multidimensional list.
            res[0] - the start of a libration period,
            res[1] - the end of a libration period,
            res[2] - the length of the given libration period.

        """
        breaks = cls.find_breaks(x, y)

        if 0 == len(breaks[1]):  # full interval is a libration
            return [[x[0]], [x[len(y) - 1]], [x[-1] - x[0]]]

        librations = [[], [], []]  # start, stop, length

        breaks_diff = np.diff(breaks[0])

        libration_start = x[0]
        libration_length = breaks[0][0] - x[0]
        prev_direction = breaks[1][0]

        if 1 == len(breaks[1]):  # pragma: no cover
            librations[0].append(breaks[0][0])
            librations[1].append(x[-1])
            librations[2].append(libration_length + (x[-1] - breaks[0][0]))
            return librations

        for i in range(1, len(breaks[0])):
            curr_direction = breaks[1][i]
            libration_length += breaks_diff[i - 1]
            if curr_direction == prev_direction:  # circulation found
                librations[0].append(libration_start)
                librations[1].append(breaks[0][i])
                librations[2].append(libration_length)

                libration_start = breaks[0][i]
                libration_length = 0.0
            if i == (len(breaks[0]) - 1):
                # If the last break has happened before the end, then add one more interval of libration.
                if breaks[0][i] == x[-1]:
                    # flush data for libration period. If there is a circulation, it is already flushed.
                    if curr_direction != prev_direction:
                        librations[0].append(libration_start)
                        librations[1].append(x[-1])
                        librations[2].append(libration_length + (x[-1] - breaks[0][i]))
                else:
                    # append last libration (because break is not in the last point)
                    librations[0].append(breaks[0][i])
                    librations[1].append(x[-1])
                    librations[2].append(libration_length + (x[-1] - breaks[0][i]))

            prev_direction = curr_direction
        return librations

    @classmethod
    def circulation_metrics(cls, librations):
        max_libration_length = max(librations[2])
        num_libration_periods = len(librations[0])

        return {'num_libration_periods': num_libration_periods, 'max_libration_length': max_libration_length}

    @classmethod
    def overlap(cls, a, b, delta=0):
        return max(0, min(a[1] + delta, b[1] + delta) - max(a[0] - delta, b[0] - delta))

    @classmethod
    def overlap_list(cls, a_list, b_list, delta=0):
        arr = []
        for a_elem in a_list:
            for b_elem in b_list:
                if cls.overlap(a_elem, b_elem, delta=delta):
                    arr.append(a_elem)
                    break
        return arr

    @classmethod
    def periodogram(cls, x, y, minimum_frequency=0.00001, maximum_frequency=0.002, nyquist_factor=5):
        """Calculates Lomb-Scargle periodogram for a time series.

        Parameters
        ----------
        x : list
            list of times
        y : list
            list of values (angles, axis e.t.c.)
        minimum_frequency : float, optional
            the minimum frequency to look for peaks, by default 0.00001
        maximum_frequency : float, optional
            the maximum frequency to look for peaks, by default 0.002
        nyquist_factor : int, optional
            the parameter from lomg-scargle method, by default 5

        Returns
        -------
        (frequence, power)
            Return list of frequencies among with related power.
        """
        frequency, power = LombScargle(x, y).autopower(
            nyquist_factor=nyquist_factor, minimum_frequency=minimum_frequency, maximum_frequency=maximum_frequency
        )
        return (frequency, power)

    @classmethod
    def find_peaks_with_position(cls, frequency, power, height=0.05, distance=10):
        peaks, props = signal.find_peaks(power, height=height, distance=distance, width=(None, None))
        peaks_right, peaks_left = (
            1.0 / frequency[np.rint(props['left_ips']).astype(int)],
            1.0 / frequency[np.rint(props['right_ips']).astype(int)],
        )
        peaks_position = list(zip(peaks_left, peaks_right))
        peaks_position = sorted(peaks_position, key=lambda tup: tup[0])
        return {'position': peaks_position, 'peaks': peaks}

    @classmethod
    def butter_lowpass_filter(cls, data, cutoff, fs, order, nyq):
        normal_cutoff = cutoff / nyq
        # Get the filter coefficients
        b, a = signal.butter(order, normal_cutoff, btype='low', analog=False)
        y = signal.filtfilt(b, a, data, method="gust")
        return y

    @classmethod
    def body(cls, sim, body: resonances.Body):

        integration_time = round(sim.tmax / (2 * np.pi))
        fs = sim.Nout / integration_time  # sample rate, Hz || Nout/time, i.e. 10000/100000
        cutoff = sim.oscillations_cutoff  # should be a little bit more than needed
        nyq = 0.5 * fs  # Nyquist Frequency
        order = sim.oscillations_filter_order  # polynom order
        """
        Do not take into account first N and last N points because of the filter applied.
        There is no previous (or following) data for them. Thus, they mess the periodogram.
        """
        points_to_cut = round(sim.libration_period_min * fs)

        axis_filtered = cls.butter_lowpass_filter(body.axis, cutoff, fs, order, nyq)
        try:
            (axis_frequency, axis_power) = resonances.libration.periodogram(
                sim.times[points_to_cut : len(axis_filtered) - points_to_cut] / (2 * np.pi),
                axis_filtered[points_to_cut : len(axis_filtered) - points_to_cut],
                minimum_frequency=sim.periodogram_frequency_min,
                maximum_frequency=sim.periodogram_frequency_max,
            )
            axis_peaks_data = cls.find_peaks_with_position(axis_frequency, axis_power, height=sim.periodogram_soft)
        except Exception as e:  # pragma: no cover
            resonances.logger.error(f"Error in periodogram of semi-major axis for {body.name}: {e}")
            axis_frequency, axis_power, axis_peaks_data = None, None, None

        body.axis_filtered = axis_filtered
        body.axis_periodogram_frequency = axis_frequency
        body.axis_periodogram_power = axis_power
        body.axis_periodogram_peaks = axis_peaks_data

        for mmr in body.mmrs:
            pure = resonances.libration.pure(body.angle(mmr))

            librations = resonances.libration.circulation(sim.times / (2 * np.pi), body.angle(mmr))
            libration_metrics = resonances.libration.circulation_metrics(librations)
            monotony = resonances.libration.monotony_estimation(body.angle(mmr))

            try:
                angle_filtered = cls.butter_lowpass_filter(body.angle(mmr), cutoff, fs, order, nyq)
                (frequency, power) = resonances.libration.periodogram(
                    sim.times[points_to_cut : len(angle_filtered) - points_to_cut] / (2 * np.pi),
                    angle_filtered[points_to_cut : len(angle_filtered) - points_to_cut],
                    minimum_frequency=sim.periodogram_frequency_min,
                    maximum_frequency=sim.periodogram_frequency_max,
                )

                angle_peaks_data = cls.find_peaks_with_position(frequency, power, height=sim.periodogram_soft)
                overlapping = cls.overlap_list(angle_peaks_data['position'], axis_peaks_data['position'], delta=0)
            except Exception as e:  # pragma: no cover
                resonances.logger.error(f"Error in periodogram for {body.name} and {mmr.to_s()}: {e}")
                frequency, power, angle_peaks_data, angle_filtered, overlapping = None, None, None, None, []

            body.statuses[mmr.to_s()] = cls.resolve(
                pure,
                overlapping,
                libration_metrics['max_libration_length'],
                sim.libration_period_critical,
                monotony,
                sim.libration_monotony_critical,
            )

            body.librations[mmr.to_s()] = librations
            body.libration_metrics[mmr.to_s()] = libration_metrics
            body.libration_pure[mmr.to_s()] = pure

            body.periodogram_frequency[mmr.to_s()] = frequency
            body.periodogram_power[mmr.to_s()] = power
            body.periodogram_peaks[mmr.to_s()] = angle_peaks_data

            body.angles_filtered[mmr.to_s()] = angle_filtered
            body.periodogram_peaks_overlapping[mmr.to_s()] = overlapping

            body.monotony[mmr.to_s()] = monotony

        return body.statuses

    @classmethod
    def resolve(cls, pure, overlapping, max_libration_length, libration_period_critical, monotony, libration_monotony_critical):
        status = 0
        if pure and (len(overlapping) > 0):
            status = 2  # pure libration
        elif pure:
            # seems to be pure but libration periods of axis and resonant angle are different
            # need manual check
            status = -2
        elif (len(overlapping) > 0) and (max_libration_length > libration_period_critical):
            status = 1  # transient resonance
        elif (
            (max_libration_length > libration_period_critical)
            and (monotony >= libration_monotony_critical[0])
            and (monotony <= libration_monotony_critical[1])
        ):
            # Looks like chaotic but has long stable period and acceptable monotony
            # need to verify manually
            status = -1
        # No resonance
        return status



import resonances
import astdys
from typing import Union, List


def convert_input_to_list(asteroids: Union[int, str, List[Union[int, str]]]) -> List[str]:
    if isinstance(asteroids, str) or isinstance(asteroids, int):
        asteroids = [asteroids]
    if (asteroids is not None) and (len(asteroids) > 0):
        asteroids = list(map(str, asteroids))
    else:
        asteroids = []
    return asteroids


def find(asteroids: Union[int, str, List[Union[int, str]]], planets=None) -> resonances.Simulation:
    sim = resonances.Simulation()
    sim.create_solar_system()

    asteroids = convert_input_to_list(asteroids)

    elems = astdys.search(asteroids)
    for asteroid in asteroids:
        elem = elems[asteroid]
        mmrs = resonances.ThreeBodyMatrix.find_resonances(elem['a'], planets=planets)
        mmrs2 = resonances.TwoBodyMatrix.find_resonances(elem['a'], planets=planets)
        mmrs = mmrs + mmrs2
        sim.add_body(elem, mmrs, f"{asteroid}")
        resonances.logger.info('Adding a possible resonance for an asteroid {} - {}'.format(asteroid, ', '.join(map(str, elems.values()))))

    # default settings
    sim.dt = 1
    sim.plot = 'nonzero'
    return sim


def check(asteroids: Union[int, str, List[Union[int, str]]], mmr: Union[resonances.MMR, str]) -> resonances.Simulation:
    if isinstance(mmr, str):
        mmr = resonances.create_mmr(mmr)

    sim = resonances.Simulation()
    sim.create_solar_system()

    asteroids = convert_input_to_list(asteroids)

    elems = astdys.search(asteroids)

    for asteroid in asteroids:
        elem = elems[asteroid]
        sim.add_body(elem, mmr, f"{asteroid}")
        resonances.logger.info('Adding a possible resonance for an asteroid {} - {}'.format(asteroid, mmr.to_s()))

    # default settings
    sim.dt = 1
    sim.plot = 'nonzero'
    return sim


def find_asteroids_in_mmr(mmr: Union[resonances.MMR, str], sigma=0.1, per_iteration=500):  # pragma: no cover
    if isinstance(mmr, str):
        mmr = resonances.create_mmr(mmr)

    df = astdys.search_by_axis(mmr.resonant_axis, sigma=sigma)
    numbers = df.index.astype(str).tolist()
    chunks = [numbers[i : i + per_iteration] for i in range(0, len(numbers), per_iteration)]

    num_chunks = len(chunks)
    data = []
    save_path = None
    plot_path = None
    for i, chunk in enumerate(chunks):
        sim = resonances.Simulation()
        sim.create_solar_system()
        sim.dt = 1
        sim.plot = 'nonzero'
        if save_path is not None:
            sim.save_path = save_path
            sim.plot_path = plot_path
        else:
            save_path = sim.save_path
            plot_path = sim.plot_path

        resonances.logger.info(f"Iteration {i+1}/{num_chunks}: Going to process a chunk of {len(chunk)} asteroids.")
        for asteroid in chunk:
            sim.add_body(df.loc[asteroid].to_dict(), mmr, f"{asteroid}")
        sim.run()
        data.append(sim.get_simulation_summary())

    return data



import numpy as np
import re
from resonances.resonance.mmr import MMR
from resonances.data import const
import rebound


class TwoBody(MMR):
    def __init__(self, coeff, planets_names=None):
        if isinstance(coeff, str):
            coeff, planets_names = self.init_from_short_notation(coeff)

        super().__init__(coeff, planets_names)

        if np.gcd(self.coeff[0], self.coeff[1]) > 1:
            raise Exception('The integers should have gcd equals 1. Given {}.'.format(', '.join(str(e) for e in coeff)))

    def calc_angle(self, body, planets):
        body1 = planets[0]
        angle = rebound.mod2pi(
            self.coeff[0] * body1.l
            + self.coeff[1] * body.l
            + self.coeff[2] * (body1.Omega + body1.omega)
            + self.coeff[3] * (body.Omega + body.omega)
        )
        return angle

    def init_from_short_notation(self, s):
        tmp = re.split('-|\\+', s)
        if 2 != len(tmp):
            raise Exception('You must specify two integers only for a short notation, i.e., 2J-1.')

        first_letter = tmp[0][len(tmp[0]) - 1]
        planets_names = [self.get_planet_name_from_letter(first_letter)]

        coeff1 = int(str.replace(tmp[0], first_letter, ''))
        coeff2 = int(tmp[1])
        symbol2 = s[len(tmp[0])]
        if symbol2 == '-':
            coeff2 = -coeff2
        coeff = [coeff1, coeff2, 0, (0 - coeff1 - coeff2)]
        return coeff, planets_names

    def to_s(self):
        s = '{:d}{:.1}{:+d}{:+d}{:+d}'.format(
            int(self.coeff[0]),
            self.get_letter_from_planet_name(self.planets_names[0]),
            int(self.coeff[1]),
            int(self.coeff[2]),
            int(self.coeff[3]),
        )
        return s

    def to_short(self):
        s = '{:d}{:.1}{:+d}'.format(
            int(self.coeff[0]),
            self.get_letter_from_planet_name(self.planets_names[0]),
            int(self.coeff[1]),
        )
        return s

    def calculate_resonant_axis(self):
        if len(self.planets_names) != 1:
            raise Exception('Cannot calculate resonant axis if the only planet is not specified!')

        planet_axis = const.PLANETS_AXIS[self.planets_names[0]]
        axis = planet_axis * (((self.coeff[1] / self.coeff[0]) ** (2.0)) ** (1.0 / 3))

        return axis



import logging
from pathlib import Path
import resonances.config


def static_init(cls):
    if getattr(cls, "static_init", None):
        cls.static_init()
    return cls


@static_init
class logger:  # pragma: no cover
    @classmethod
    def static_init(cls):
        log_file_path = resonances.config.get('log.file')
        log_dir = Path(log_file_path).parent.resolve()
        Path(log_dir).mkdir(parents=True, exist_ok=True)
        logging.basicConfig(
            filename=log_file_path,
            level=cls.get_logging_level(),
            format='%(asctime)s %(levelname)s: %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
        )

    @classmethod
    def get_logging_level(cls):
        config_level = resonances.config.get('log.level')
        if 'debug' == config_level:
            return logging.DEBUG
        elif 'warning' == config_level:
            return logging.WARNING
        elif 'error' == config_level:
            return logging.ERROR
        elif 'critical' == config_level:
            return logging.CRITICAL

        return logging.INFO

    @classmethod
    def info(cls, message):
        logging.info(message)

    @classmethod
    def debug(cls, message):
        logging.debug(message)

    @classmethod
    def warning(cls, message):
        logging.warning(message)

    @classmethod
    def error(cls, message):
        logging.error(message)



import numpy as np
from .resonance.mmr import MMR
from .logger import logger
from typing import List


class Body:
    def __init__(self, type='particle'):
        self.type = type

        self.initial_data = None
        self.name = ''
        self.mass = 0.0

        # Integration data
        self.times = None
        self.axis = None
        self.ecc = None
        self.longitude = None
        self.varpi = None

        # resonances data
        # keys are resonances (4J-2S-1-0-0-1), values are values

        self.mmrs: List[MMR] = []
        self.angles = {}
        self.statuses = {}

        # Libration and filtering data
        self.librations = {}
        self.libration_metrics = {}
        self.libration_status = {}
        self.libration_pure = {}

        self.periodogram_frequency = {}
        self.periodogram_power = {}
        self.periodogram_peaks = {}

        self.angles_filtered = {}

        self.axis_filtered = None
        self.axis_periodogram_frequency = None
        self.axis_periodogram_power = None
        self.axis_periodogram_peaks = None

        self.periodogram_peaks_overlapping = {}

        self.monotony = {}

        # Simulation data
        self.index_in_simulation = None
        # self.index_of_planets = None

    def __str__(self):
        s = f'Body(type={self.type}, name={self.name}, mass={self.mass})\n Resonances: '
        for mmr in self.mmrs:
            s += mmr.to_s() + ', '
        return s

    def mmr_to_dict(self, mmr: MMR, times: np.ndarray):
        try:
            df_data = {
                'times': times / (2 * np.pi),
                'angle': self.angles[mmr.to_s()],
                'a': self.axis,
                'e': self.ecc,
            }
            if self.periodogram_power is not None:
                len_diff = len(self.angles[mmr.to_s()]) - len(self.periodogram_power[mmr.to_s()])
                df_data['periodogram'] = np.append(self.periodogram_power[mmr.to_s()], np.zeros(len_diff))
                df_data['a_filtered'] = self.axis_filtered
                df_data['a_periodogram'] = np.append(self.axis_periodogram_power, np.zeros(len_diff))
        except Exception as e:
            logger.error(f'Error in mmr_to_dict for body={self.name} and mmr={mmr.to_s()}: {e}')
            return None
        return df_data

    def setup_vars_for_simulation(self, num):
        self.axis, self.ecc, self.longitude, self.varpi = (
            np.zeros(num),
            np.zeros(num),
            np.zeros(num),
            np.zeros(num),
        )
        for mmr in self.mmrs:
            self.angles[mmr.to_s()] = np.zeros(num)

    def angle(self, mmr: MMR) -> np.ndarray:
        try:
            return self.angles[mmr.to_s()]
        except Exception:
            raise Exception('The angle for the resonance {} does not exist in the body {}.'.format(mmr.to_s(), self.name))

    def in_resonance(self, mmr: MMR):
        if (self.status(mmr) is not None) and (self.status(mmr) > 0):
            return True
        return False

    def status(self, mmr: MMR):
        return self.statuses[mmr.to_s()]

    def in_pure_resonance(self, mmr: MMR):
        if (self.status(mmr) is not None) and (2 == self.status(mmr)):
            return True
        return False

    def is_particle(self):
        if 'particle' == self.type:
            return True
        return False



__version__ = '0.1.1'
import logging

from .config import config
from .logger import logger

from .resonance.three_body import ThreeBody
from .resonance.two_body import TwoBody
from .resonance.mmr import MMR
from .resonance.factory import create_mmr
from resonances.matrix.three_body_matrix import ThreeBodyMatrix
from resonances.matrix.two_body_matrix import TwoBodyMatrix
from resonances.body import Body
from .simulation import Simulation

from .resonance.libration import libration

import resonances.resonance.plot

# import resonances.data.const
from resonances.finder import find
from resonances.finder import check
from resonances.finder import find_asteroids_in_mmr



import numpy as np
import pandas as pd

import resonances
from resonances.matrix.matrix import Matrix


class TwoBodyMatrix(Matrix):
    catalog_file = 'matrix.2body.file'

    @classmethod
    def build(cls):
        primary_max = resonances.config.get('matrix.2body.primary_max')
        m_max = resonances.config.get('matrix.2body.coefficients_max')
        q_max = resonances.config.get('matrix.2body.max_order')
        if (cls.planets is None) or (len(cls.planets) == 0):
            planets = resonances.data.const.SOLAR_SYSTEM
        else:
            planets = cls.planets
        data = []
        for planet in planets:
            for m1 in range(1, primary_max):
                for m in range(-m_max, m_max):
                    if np.gcd(m1, m) > 1:
                        continue
                    if m == 0:
                        continue
                    p = 0 - (m1 + m)
                    if abs(p) > q_max:
                        continue
                    mmr = resonances.TwoBody([m1, m, 0, p], [planet])
                    try:
                        axis = mmr.resonant_axis
                        data.append([mmr.to_short(), planet, m1, m, abs(p), axis])
                    except Exception:  # pragma: no cover
                        continue

        df = pd.DataFrame(data, columns=['mmr', 'planet', 'm1', 'm', 'q', 'a'])
        cls.matrix = df
        return df

    @classmethod
    def find_resonances(cls, a, sigma=0.1, planets=None):
        if cls.matrix is None:
            cls.load()

        if isinstance(planets, list):
            df = cls.matrix[(cls.matrix['a'] >= (a - sigma)) & (cls.matrix['a'] <= (a + sigma)) & (cls.matrix['planet'].isin(planets))]
        else:
            df = cls.matrix[(cls.matrix['a'] >= (a - sigma)) & (cls.matrix['a'] <= (a + sigma))]

        mmrs = []
        for mmr in df['mmr'].tolist():
            mmrs.append(resonances.create_mmr(mmr))
        return mmrs



import numpy as np


class MMR:

    def __init__(self, coeff, planets_names=None, index_of_planets=None):
        self.coeff = np.array(coeff)
        if self.coeff[0] <= 0:
            raise Exception("The primary coefficient of the resonance should be greater than 0. Given {}.".format(self.coeff[0]))

        if sum(self.coeff) != 0:
            raise Exception(
                "Sum of integers in a resonance should follow the D'Alembert rule. Given {}, the sum is equal to {}.".format(
                    ', '.join(str(e) for e in coeff), sum(self.coeff)
                )
            )

        if planets_names is None:
            self.planets_names = []
        else:
            self.planets_names = planets_names

        self._resonant_axis = None
        self.index_of_planets = index_of_planets

    def __str__(self):
        return "MMR(coeff=[{}])".format(', '.join(str(e) for e in self.coeff))

    def number_of_bodies(self):
        return len(self.coeff) / 2

    def get_planet_name_from_letter(self, letter):
        if letter == 'R':
            return 'Mercury'
        elif letter == 'V':
            return 'Venus'
        elif letter == 'E':
            return 'Earth'
        elif letter == 'M':
            return 'Mars'
        elif letter == 'J':
            return 'Jupiter'
        elif letter == 'S':
            return 'Saturn'
        elif letter == 'U':
            return 'Uranus'
        elif letter == 'N':
            return 'Neptune'
        raise Exception('Bad notation used. Only the following letter are available: R (for Mercury), V, E, M, J, S, U, N ')

    def get_letter_from_planet_name(self, planet_name: str) -> str:
        if 'Mercury' == planet_name:
            return 'R'
        return planet_name[0]

    @property
    def resonant_axis(self):
        if self._resonant_axis is None:
            self._resonant_axis = self.calculate_resonant_axis()
        return self._resonant_axis

    @resonant_axis.setter
    def resonant_axis(self, value):
        self._resonant_axis = value

    def calculate_resonant_axis(self):  # pragma: no cover
        pass



from typing import List
import numpy as np
import pandas as pd
import itertools

import resonances
from resonances.matrix.matrix import Matrix


class ThreeBodyMatrix(Matrix):

    catalog_file = 'matrix.3body.file'

    @classmethod
    def build(cls):
        primary_max = resonances.config.get('matrix.3body.primary_max')
        m_max = resonances.config.get('matrix.3body.coefficients_max')
        q_max = resonances.config.get('matrix.3body.max_order')
        if (cls.planets is None) or (len(cls.planets) == 0):
            planets = resonances.data.const.SOLAR_SYSTEM
        else:
            planets = cls.planets
        pairs = list(itertools.combinations(planets, 2))
        data = []
        for planet1, planet2 in pairs:
            for m1 in range(1, primary_max):
                for m2 in range(-m_max, m_max):
                    for m in range(-m_max, m_max):
                        if np.gcd(np.gcd(m1, m2), m) > 1:
                            continue
                        if (m2 == 0) or (m == 0):
                            continue
                        p = 0 - (m1 + m2 + m)
                        if abs(p) > q_max:
                            continue
                        mmr = resonances.ThreeBody([m1, m2, m, 0, 0, p], [planet1, planet2])
                        try:
                            axis = mmr.resonant_axis
                            data.append([mmr.to_short(), planet1, planet2, m1, m2, m, abs(p), axis])
                        except Exception:
                            continue

        df = pd.DataFrame(data, columns=['mmr', 'planet1', 'planet2', 'm1', 'm2', 'm', 'q', 'a'])
        cls.matrix = df
        return df

    @classmethod
    def find_resonances(cls, a, sigma=0.02, planets=None) -> List[resonances.MMR]:
        if cls.matrix is None:
            cls.load()

        if isinstance(planets, list):
            df = cls.matrix[
                (cls.matrix['a'] >= (a - sigma))
                & (cls.matrix['a'] <= (a + sigma))
                & (cls.matrix['planet1'].isin(planets))
                & (cls.matrix['planet2'].isin(planets))
            ]
        else:
            df = cls.matrix[(cls.matrix['a'] >= (a - sigma)) & (cls.matrix['a'] <= (a + sigma))]

        mmrs = []
        for mmr in df['mmr'].tolist():
            mmrs.append(resonances.create_mmr(mmr))
        return mmrs



import numpy as np
import re
from resonances.resonance.mmr import MMR
import resonances.data.util as util
from resonances.data import const
import rebound


class ThreeBody(MMR):
    def __init__(self, coeff, planets_names=None):
        if isinstance(coeff, str):
            coeff, planets_names = self.init_from_short_notation(coeff)

        super().__init__(coeff, planets_names)

        if np.gcd(np.gcd(self.coeff[0], self.coeff[1]), self.coeff[2]) > 1:
            raise Exception('The integers should have gcd equals 1. Given {}.'.format(', '.join(str(e) for e in coeff)))

    def calc_angle(self, body, planets):
        body1 = planets[0]
        body2 = planets[1]
        angle = rebound.mod2pi(
            self.coeff[0] * body1.l
            + self.coeff[1] * body2.l
            + self.coeff[2] * body.l
            + self.coeff[3] * (body1.Omega + body1.omega)
            + self.coeff[4] * (body2.Omega + body2.omega)
            + self.coeff[5] * (body.Omega + body.omega)
        )
        return angle

    def init_from_short_notation(self, s):
        tmp = re.split('-|\\+', s)
        if 3 != len(tmp):
            raise Exception('You must specify three integers only for a short notation, i.e., 4J-2S-1.')

        first_letter = tmp[0][len(tmp[0]) - 1]
        second_letter = tmp[1][len(tmp[1]) - 1]
        planets_names = [self.get_planet_name_from_letter(first_letter), self.get_planet_name_from_letter(second_letter)]

        coeff1 = int(str.replace(tmp[0], first_letter, ''))
        coeff2 = int(str.replace(tmp[1], second_letter, ''))
        coeff3 = int(tmp[2])
        symbol2 = s[len(tmp[0])]
        symbol3 = s[len(tmp[0]) + len(tmp[1]) + 1]
        if symbol2 == '-':
            coeff2 = -coeff2
        if symbol3 == '-':
            coeff3 = -coeff3
        coeff = [coeff1, coeff2, coeff3, 0, 0, (0 - coeff1 - coeff2 - coeff3)]
        return coeff, planets_names

    def to_s(self):
        s = '{:d}{:.1}{:+d}{:.1}{:+d}{:+d}{:+d}{:+d}'.format(
            int(self.coeff[0]),
            self.get_letter_from_planet_name(self.planets_names[0]),
            int(self.coeff[1]),
            self.get_letter_from_planet_name(self.planets_names[1]),
            int(self.coeff[2]),
            int(self.coeff[3]),
            int(self.coeff[4]),
            int(self.coeff[5]),
        )
        return s

    def to_short(self):
        s = '{:d}{:.1}{:+d}{:.1}{:+d}'.format(
            int(self.coeff[0]),
            self.get_letter_from_planet_name(self.planets_names[0]),
            int(self.coeff[1]),
            self.get_letter_from_planet_name(self.planets_names[1]),
            int(self.coeff[2]),
        )
        return s

    @property
    def resonant_axis(self):
        if self._resonant_axis is None:
            self._resonant_axis = self.calculate_resonant_axis()
        return self._resonant_axis

    @resonant_axis.setter
    def resonant_axis(self, value):
        self._resonant_axis = value

    def calculate_resonant_axis(self):
        if len(self.planets_names) != 2:
            raise Exception('Cannot calculate resonant axis if the amount of planets is not equal to 2!')

        """ "
        a_i - semi-major axis of two planets and the body
        n_i — mean motions
        l_i — longitude of node
        p_i — the names of the planets
        The indexes 1,2 refers to the planets, no index — to the body
        k — gravitational constant (~0.017)
        """
        p1 = self.planets_names[0]
        p2 = self.planets_names[1]

        n1 = util.mean_motion_from_axis(const.PLANETS_AXIS[p1])
        n2 = util.mean_motion_from_axis(const.PLANETS_AXIS[p2])

        l1 = 0
        l2 = 0

        n = (-self.coeff[0] * n1 - self.coeff[1] * n2 - self.coeff[3] * l1 - self.coeff[4] * l2) / self.coeff[2]

        if n < 0:
            raise Exception('Something weird has happened for {}. Mean motion = {}'.format(self.to_short(), n))

        la = 0.0
        for planet in const.SOLAR_SYSTEM:
            ap = const.PLANETS_AXIS[planet]
            mp = const.PLANETS_MASS[planet]
            a = util.axis_from_mean_motion(n)

            if a > ap:
                la = la + (3 * np.pi / 2 * mp / (1.0 + mp) ** 1.5 * (ap**2 / a**3.5)) / const.DAYS_IN_YEAR
            else:
                la = la + (3 * np.pi / 2 * mp * (np.sqrt(a) / ap) ** 3) / const.DAYS_IN_YEAR

        n = (-self.coeff[0] * n1 - self.coeff[1] * n2 - self.coeff[3] * l1 - self.coeff[4] * l2 - self.coeff[5] * la) / self.coeff[2]

        if n < 0:  # pragma: no cover
            raise Exception('Something weird has happened for {}. Mean motion = {}'.format(self.to_short(), n))

        a = util.axis_from_mean_motion(n)

        return a

    # Yes, I know that it is a bad practice to keep commented code.
    # However, this might be useful for some edge cases.
    # def another_calculate_axis(self):
    #     """ "
    #     a_i - semi-major axis of two planets and the body
    #     n_i — mean motions
    #     l_i — longitude of node
    #     p_i — the names of the planets
    #     The indexes 1,2 refers to the planets, no index — to the body
    #     k — gravitational constant (~0.017)
    #     """
    #     p1 = self.planets_names[0]
    #     p2 = self.planets_names[1]

    #     a1 = const.PLANETS_AXIS[p1]
    #     a2 = const.PLANETS_AXIS[p2]

    #     n1 = util.mean_motion_from_axis(a1)
    #     n2 = util.mean_motion_from_axis(a2)

    #     l1 = util.from_s_to_yr(const.PLANETS_LONGITUDE[p1])
    #     l2 = util.from_s_to_yr(const.PLANETS_LONGITUDE[p2])

    #     k = const.K

    #     n = (-self.coeff[0] * n1 - self.coeff[1] * n2 - self.coeff[3] * l1 - self.coeff[4] * l2) / self.coeff[2]

    #     if n < 0:
    #         raise Exception('Something weird has happened for {}. Mean motion = {}'.format(self.to_short(), n))

    #     a = util.axis_from_mean_motion(n)
    #     eps = (a1 - a) / a1
    #     l = k / (2 * np.pi) * np.sqrt(a / a1) * (eps ** 2) * n1

    #     n = (-self.coeff[0] * n1 - self.coeff[1] * n2 - self.coeff[3] * l1 - self.coeff[4] * l2 - self.coeff[5] * l) / self.coeff[2]

    #     if n < 0:
    #         raise Exception('Something weird has happened for {}. Mean motion = {}'.format(self.to_short(), n))

    #     a = util.axis_from_mean_motion(n)
    #     return a



import matplotlib.pyplot as plt
import numpy as np

import resonances.config
import resonances


def body(sim, body: resonances.Body, mmr: resonances.MMR, image_type='png'):
    plt.style.use('default')

    fig, axs = plt.subplots(6, 1, figsize=(10, 10))

    fig.suptitle(
        "{}, resonance = {}, status = {}".format(body.name, mmr.to_short(), body.statuses[mmr.to_s()]),
        fontsize=14,
    )

    axs[0].set_title('Resonant angle')
    axs[0].set_xlim([0, sim.tmax_yrs])
    axs[0].xaxis.set_major_locator(plt.MultipleLocator(10000))
    axs[0].xaxis.set_minor_locator(plt.MultipleLocator(2000))
    axs[0].plot(sim.times / (2 * np.pi), body.angle(mmr), linestyle='', marker=',')

    if body.angles_filtered[mmr.to_s()] is not None:  # pragma: no cover
        axs[1].set_title('Filtered resonant angle')
        axs[1].plot(sim.times / (2 * np.pi), body.angles_filtered[mmr.to_s()], linestyle='', marker=',')
    else:
        axs[1].set_title('Again resonant angle (no filtered available)')
        axs[1].plot(sim.times / (2 * np.pi), body.angle(mmr), linestyle='', marker=',')
    axs[1].sharex(axs[0])

    if body.axis_filtered is not None:  # pragma: no cover
        axs[2].set_title('Filtered semi-major axis')
        axs[2].plot(sim.times / (2 * np.pi), body.axis_filtered, linestyle='', marker=',')
    else:
        axs[2].set_title('Semi-major axis')
        axs[2].plot(sim.times / (2 * np.pi), body.axis, linestyle='', marker=',')
    axs[2].sharex(axs[0])

    axs[3].set_xlim(0, 40000)
    # axs[2].set_ylim(0, 0.2)
    axs[3].axhline(y=0.05, color='r', linestyle='--')
    axs[3].axhline(y=0.1, color='g', linestyle='--')
    if (
        (body.periodogram_peaks[mmr.to_s()] is not None)
        and ('peaks' in body.periodogram_peaks[mmr.to_s()])
        and (body.periodogram_peaks[mmr.to_s()]['peaks'].size)
    ):  # pragma: no cover
        peaks = body.periodogram_peaks[mmr.to_s()]['peaks']
        for peak_width in body.periodogram_peaks[mmr.to_s()]['position']:
            axs[3].axvline(x=peak_width[0], color='gray', linestyle="dashed")
            axs[3].axvline(x=peak_width[1], color='gray', linestyle='--')
        axs[3].plot(1.0 / body.periodogram_frequency[mmr.to_s()][peaks], body.periodogram_power[mmr.to_s()][peaks], 'x', color='orange')
        axs[3].plot(1.0 / body.periodogram_frequency[mmr.to_s()], body.periodogram_power[mmr.to_s()])

    axs[3].set_title('Periodogram (semi-major axis)')
    axs[4].set_title('Periodogram (the resonant angle)')
    axs[4].set_xlim(0, 40000)
    # axs[3].set_ylim(0, 0.2)
    axs[4].axhline(y=0.05, color='r', linestyle='--')
    axs[4].axhline(y=0.1, color='g', linestyle='--')
    if (
        (body.axis_periodogram_peaks is not None)
        and ('peaks' in body.axis_periodogram_peaks)
        and (body.axis_periodogram_peaks['peaks'].size)
    ):  # pragma: no cover
        peaks = body.axis_periodogram_peaks['peaks']
        for peak_width in body.axis_periodogram_peaks['position']:
            axs[4].axvline(x=peak_width[0], color='gray', linestyle="dashed")
            axs[4].axvline(x=peak_width[1], color='gray', linestyle='--')
        axs[4].plot(1.0 / body.axis_periodogram_frequency[peaks], body.axis_periodogram_power[peaks], 'x', color='orange')
        axs[4].plot(1.0 / body.axis_periodogram_frequency, body.axis_periodogram_power)

        axs[4].sharex(axs[3])

    axs[5].set_title('Eccentricity')
    axs[5].plot(sim.times / (2 * np.pi), body.ecc, linestyle='', marker=',')
    axs[5].sharex(axs[0])

    plt.tight_layout()

    if sim.plot_type in ['both', 'save']:
        plt.savefig('{}/{}_{}.{}'.format(sim.plot_path, body.name, mmr.to_s(), image_type))

    if sim.plot_type in ['both', 'show']:  # pragma: no cover
        plt.show()
